#!/usr/bin/python3
import io
import sys
import xml.etree.ElementTree as ET
from typing import Iterator, Optional, TextIO, Tuple, Union


class BadReport(Exception):
  pass


class HlsResources:
  """An object representing the HLS resource estimation.

  This class supports add and sub between the same class and mul with a number.
  URAM is not supported yet.

  Attributes:
    RESOURCES: Static, resource attribute names expected.
    name: Optional name of the module.
  """
  RESOURCES = 'FF', 'LUT', 'BRAM_18K', 'DSP48E'

  def __init__(self,
               obj: Union['HlsResources', ET.Element, TextIO, None] = None
              ) -> None:
    """Initialize an HlsResources from another HlsResources, a report or None.

    If obj is another HlsResources, it will be copied.
    If obj is an xml.etree.ElementTree.Element, it has to be pointing to a
      valid HLS report XML tree.
    If obj is an io.TextIOBase, it will be parsed as a valid HLS report XML.
    If obj is None, all resources will be 0.

    Args:
      obj: Object used for initialization.

    Raises:
      TypeError: If obj is not of a correct type.
    """
    self.name = None  # type: Optional[str]
    if isinstance(obj, HlsResources):
      self.name = obj.name
      for resource in HlsResources.RESOURCES:
        self[resource] = obj[resource]
      return
    if isinstance(obj, ET.Element):
      self.init_from_xml_element(obj)
      return
    if isinstance(obj, io.TextIOBase):
      self.init_from_xml_element(ET.parse(obj).getroot())
      return
    if obj is None:
      for resource in HlsResources.RESOURCES:
        self[resource] = 0
      return
    raise TypeError('obj must be '
                    'an xml.etree.ElementTree.Element object or '
                    'another HlsResources object or '
                    'an io.TextIOBase object or '
                    'None')

  def init_from_xml_element(self, elem: ET.Element,
                            item: str = 'Resources') -> 'HlsResources':
    """Initialize self from an xml.etree.ElementTree.Element.

    Initialize self from an xml.etree.ElementTree.Element with the option of
    reading the used resources or the available resources.

    Args:
      elem: Must be an xml.etree.ElementTree.Element pointing to a valid HLS
      report XML tree.
      item: String specifying what to read. Valid values are 'Resources' and
        'AvailableResources'.

    Returns:
      Updated self.
    """
    if item == 'Resources':
      self.name = elem.findtext('UserAssignments/TopModelName')
    for resource in HlsResources.RESOURCES:
      self[resource] = int(
          elem.findtext('/'.join(('AreaEstimates', item, resource))))
    return self

  def __getitem__(self, key: str) -> int:
    if key in HlsResources.RESOURCES:
      return getattr(self, key)
    raise ValueError('invalid key: %s' % key)

  def __setitem__(self, key: str, value: int) -> int:
    if key in HlsResources.RESOURCES:
      setattr(self, key, value)
      return value
    raise ValueError('invalid key: %s' % key)

  def __iter__(self) -> Iterator[Tuple[str, int]]:
    return ((r, self[r]) for r in HlsResources.RESOURCES)

  def __str__(self) -> str:
    return 'module: %16s, %s' % (self.name or 'n/a', ', '.join(
        '%s: %7d' % item for item in self))

  def __add__(self, other: 'HlsResources') -> 'HlsResources':
    result = HlsResources(self)
    for resource in HlsResources.RESOURCES:
      result[resource] += other[resource]
    return result

  def __sub__(self, other: 'HlsResources') -> 'HlsResources':
    result = HlsResources(self)
    for resource in HlsResources.RESOURCES:
      result[resource] -= other[resource]
    return result

  def __mul__(self, other: float) -> 'HlsResources':
    result = HlsResources(self)
    for resource in HlsResources.RESOURCES:
      result[resource] = int(result[resource] * other)
    return result

  def __eq__(self, other) -> bool:
    if not isinstance(other, HlsResources):
      raise NotImplementedError
    for resource in HlsResources.RESOURCES:
      if self[resource] != other[resource]:
        return False
    return True

  def __hash__(self) -> int:
    return hash(tuple(self[r] for r in HlsResources.RESOURCES))


class HlsPerformance:
  """An object representing the HLS performance estimation.

  Attributes:
    name: Optional name of the module.
    ii: Integer of pipeline II.
    depth: Integer of pipeline depth.
  """

  def __init__(self,
               obj: Union['HlsPerformance', ET.Element, TextIO, None] = None
              ) -> None:
    """Initialize an HlsPerformance from another HlsPerformance, a report, or
    None.

    If obj is another HlsPerformance, it will be copied.
    If obj is an xml.etree.ElementTree.Element, it has to be pointing to a
      valid HLS report XML tree.
    If obj is an io.TextIOBase, it will be parsed as a valid HLS report XML.
    If obj is None, the ii and depth will be 0.

    Args:
      obj: Object used for initialization.

    Raises:
      TypeError: If obj is not of a correct type.
    """
    self.name, self.ii, self.depth = None, 0, 0
    if isinstance(obj, HlsPerformance):
      self.name, self.ii, self.depth = obj.name, obj.ii, obj.depth
      return
    if isinstance(obj, ET.Element):
      self.init_from_xml_element(obj)
      return
    if isinstance(obj, io.TextIOBase):
      self.init_from_xml_element(ET.parse(obj).getroot())
      return
    if obj is None:
      return
    raise TypeError('obj must be '
                    'an xml.etree.ElementTree.Element object or '
                    'another HlsPerformances object or '
                    'an io.TextIOBase object or '
                    'None')

  def init_from_xml_element(self, elem: ET.Element) -> 'HlsPerformance':
    """Initialize self from an xml.etree.ElementTree.Element.

    Initialize self from an xml.etree.ElementTree.Element.

    Args:
      elem: Must be an xml.etree.ElementTree.Element pointing to a valid HLS
      report XML tree.

    Returns:
      Updated self.
    """

    def raise_if_not_found(item: ET.Element, key: str) -> int:
      val = item.findtext(key)
      if val is None:
        raise BadReport('cannot find %s')
      return int(val)

    for item in elem.findall('PerformanceEstimates/SummaryOfLoopLatency/*'):
      self.ii = raise_if_not_found(item, 'PipelineII')
      self.depth = raise_if_not_found(item, 'PipelineDepth')
    return self


def resources(obj: Union[TextIO, str]) -> HlsResources:
  """Read HLS resource estimation from HLS report XML.

  Args:
    obj: A file-like object or a file path of HLS report XML file, which is used
      to parse the resource usage.

  Returns:
    The HLS resource estimation.

  Raises:
    TypeError: If obj is not a text file-like object nor a path as str.
  """
  if isinstance(obj, io.TextIOBase):
    return HlsResources(obj)
  if isinstance(obj, str):
    with open(obj, 'r') as fd:
      result = resources(fd)
    return result
  raise TypeError('obj must be a text file-like object or a path as str')


def performance(obj: Union[TextIO, str]) -> HlsPerformance:
  """Read HLS performance estimation from HLS report XML.

  Args:
    obj: A file-like object or a file path of HLS report XML file, which is used
      to parse the performance estimation.

  Returns:
    The HLS performance estimation.

  Raises:
    TypeError: If obj is not a text file-like object nor a path as str.
  """
  if isinstance(obj, io.TextIOBase):
    return HlsPerformance(obj)
  if isinstance(obj, str):
    with open(obj, 'r') as fd:
      result = performance(fd)
    return result
  raise TypeError('obj must be a text file-like object or a path as str')


def available_resources(obj: Union[TextIO, str]) -> HlsResources:
  """Read available resources from HLS report XML.

  Args:
    obj: A file-like object or a file path of HLS report XML file, which is used
      to parse the available resource.

  Returns:
    The HLS available resource.

  Raises:
    TypeError: If obj is not a text file-like object nor a path as str.
  """
  if isinstance(obj, io.TextIOBase):
    return HlsResources().init_from_xml_element(
        ET.parse(obj).getroot(), 'AvailableResources')
  if isinstance(obj, str):
    with open(obj, 'r') as fd:
      result = available_resources(fd)
    return result
  raise TypeError('obj must be a text file-like object or a path as str')


def main() -> None:
  if sys.argv[1:]:
    total = available_resources(sys.argv[1])
    total_used = HlsResources()
    for f in sys.argv[1:]:
      used = resources(f)
      total_used += used
      print('     used:', used)
    print('    total:', total_used)
    print('available:', total)


if __name__ == '__main__':
  main()



import collections
import contextlib
import glob
import logging
import os
import subprocess
import tarfile
import tempfile
import xml.etree.ElementTree as ET
import xml.sax.saxutils
import zipfile
from typing import (BinaryIO, Iterable, Iterator, Mapping, Optional, TextIO,
                    Tuple, Union)

from haoda import ir, util

_logger = logging.getLogger().getChild(__name__)


class Vivado(subprocess.Popen):
  """Call vivado with the given Tcl commands and arguments.

  This is a subclass of subprocess.Popen. A temporary directory will be created
  and used as the working directory.

  Args:
    commands: A string of Tcl commands.
    args: Iterable of strings as arguments to the Tcl commands.
  """

  def __init__(self, commands: str, *args: Iterable[str]):
    self.cwd = tempfile.TemporaryDirectory(prefix='vivado-')
    with open(os.path.join(self.cwd.name, 'commands.tcl'),
              mode='w+') as tcl_file:
      tcl_file.write(commands)
    cmd_args = [
        'vivado', '-mode', 'batch', '-source', tcl_file.name, '-nojournal',
        '-tclargs', *args
    ]
    pipe_args = {'stdout': subprocess.PIPE, 'stderr': subprocess.PIPE}
    super().__init__(cmd_args, cwd=self.cwd.name, **pipe_args)  # type: ignore

  def __exit__(self, *args) -> None:
    super().__exit__(*args)
    self.cwd.cleanup()


class VivadoHls(subprocess.Popen):
  """Call vivado_hls with the given Tcl commands.

  This is a subclass of subprocess.Popen. A temporary directory will be created
  and used as the working directory.

  Args:
    commands: A string of Tcl commands.
  """

  def __init__(self, commands: str):
    self.cwd = tempfile.TemporaryDirectory(prefix='vivado-hls-')
    with open(os.path.join(self.cwd.name, 'commands.tcl'),
              mode='w+') as tcl_file:
      tcl_file.write(commands)
    cmd_args = ['vivado_hls', '-f', tcl_file.name]
    pipe_args = {'stdout': subprocess.PIPE, 'stderr': subprocess.PIPE}
    super().__init__(cmd_args, cwd=self.cwd.name, **pipe_args)  # type: ignore

  def __exit__(self, *args) -> None:
    super().__exit__(*args)
    self.cwd.cleanup()


PACKAGEXO_COMMANDS = r'''
set tmp_ip_dir "{tmpdir}/tmp_ip_dir"
set tmp_project "{tmpdir}/tmp_project"

create_project -force kernel_pack ${{tmp_project}}
add_files -norecurse [glob {hdl_dir}/*.v]
foreach tcl_file [glob -nocomplain {hdl_dir}/*.tcl] {{
  source ${{tcl_file}}
}}
set_property top {top_name} [current_fileset]
update_compile_order -fileset sources_1
update_compile_order -fileset sim_1
ipx::package_project -root_dir ${{tmp_ip_dir}} -vendor xilinx.com -library RTLKernel -taxonomy /KernelIP -import_files -set_current false
ipx::unload_core ${{tmp_ip_dir}}/component.xml
ipx::edit_ip_in_project -upgrade true -name tmp_edit_project -directory ${{tmp_ip_dir}} ${{tmp_ip_dir}}/component.xml
set_property core_revision 2 [ipx::current_core]
foreach up [ipx::get_user_parameters] {{
  ipx::remove_user_parameter [get_property NAME ${{up}}] [ipx::current_core]
}}
set_property sdx_kernel true [ipx::current_core]
set_property sdx_kernel_type rtl [ipx::current_core]
ipx::create_xgui_files [ipx::current_core]
{bus_ifaces}
set_property xpm_libraries {{XPM_CDC XPM_MEMORY XPM_FIFO}} [ipx::current_core]
set_property supported_families {{ }} [ipx::current_core]
set_property auto_family_support_level level_2 [ipx::current_core]
ipx::update_checksums [ipx::current_core]
ipx::save_core [ipx::current_core]
close_project -delete

package_xo -force -xo_path "{xo_file}" -kernel_name {top_name} -ip_directory ${{tmp_ip_dir}} -kernel_xml {kernel_xml}{cpp_kernels}
'''

BUS_IFACE = r'''
ipx::associate_bus_interfaces -busif {} -clock ap_clk [ipx::current_core]
'''


class PackageXo(Vivado):
  """Packages the given files into a Xilinx hardware object.

  This is a subclass of subprocess.Popen. A temporary directory will be created
  and used as the working directory.

  Args:
    xo_file: Name of the generated xo file.
    top_name: Top-level module name.
    kernel_xml: Name of a xml file containing description of the kernel.
    hdl_dir: Directory name containing all HDL files.
    m_axi_names: Variable names connected to the m_axi bus.
    iface_names: Other interface names, default to ('s_axi_control').
    cpp_kernels: File names of C++ kernels.
  """

  def __init__(self,
               xo_file: str,
               top_name: str,
               kernel_xml: str,
               hdl_dir: str,
               m_axi_names: Iterable[str] = (),
               iface_names: Iterable[str] = ('s_axi_control',),
               cpp_kernels=()):
    self.tmpdir = tempfile.TemporaryDirectory(prefix='package-xo-')
    if _logger.isEnabledFor(logging.INFO):
      for _, _, files in os.walk(hdl_dir):
        for filename in files:
          _logger.info('packing: %s', filename)
    iface_names = list(iface_names)
    iface_names.extend(map('m_axi_{}'.format, m_axi_names))
    kwargs = {
        'top_name': top_name,
        'kernel_xml': kernel_xml,
        'hdl_dir': hdl_dir,
        'xo_file': xo_file,
        'bus_ifaces': ''.join(map(BUS_IFACE.format, iface_names)),
        'tmpdir': self.tmpdir.name,
        'cpp_kernels': ''.join(map(' -kernel_files {}'.format, cpp_kernels))
    }
    super().__init__(PACKAGEXO_COMMANDS.format(**kwargs))

  def __exit__(self, *args) -> None:
    super().__exit__(*args)
    self.tmpdir.cleanup()


HLS_COMMANDS = r'''
cd "{project_dir}"
open_project "{project_name}"
set_top {top_name}
{add_kernels}
open_solution "{solution_name}"
set_part {{{part_num}}}
create_clock -period {clock_period} -name default
config_compile -name_max_length 253
config_interface -m_axi_addr64
config_rtl -disable_start_propagation -reset_level {reset_level}
csynth_design
exit
'''


class RunHls(VivadoHls):
  """Runs Vivado HLS for the given kernels and generate HDL files

  This is a subclass of subprocess.Popen. A temporary directory will be created
  and used as the working directory.

  Args:
    tarfileobj: File object that will contain the reports and HDL files.
    kernel_files: File names or tuple of file names and cflags of the kernels.
    top_name: Top-level module name.
    clock_period: Target clock period.
    part_num: Target part number.
  """

  def __init__(self,
               tarfileobj: BinaryIO,
               kernel_files: Iterable[Union[str, Tuple[str, str]]],
               top_name: str,
               clock_period: str,
               part_num: str,
               reset_low: bool = True):
    self.project_dir = tempfile.TemporaryDirectory(prefix='run-hls-')
    self.project_name = 'project'
    self.solution_name = top_name
    self.tarfileobj = tarfileobj
    kernels = []
    for kernel_file in kernel_files:
      if isinstance(kernel_file, str):
        kernels.append(
            'add_files "{}" -cflags "-std=c++11"'.format(kernel_file))
      else:
        kernels.append(
            'add_files "{}" -cflags "-std=c++11 {}"'.format(*kernel_file))
    kwargs = {
        'project_dir': self.project_dir.name,
        'project_name': self.project_name,
        'solution_name': self.solution_name,
        'top_name': top_name,
        'add_kernels': '\n'.join(kernels),
        'part_num': part_num,
        'clock_period': clock_period,
        'reset_level': 'low' if reset_low else 'high',
    }
    super().__init__(HLS_COMMANDS.format(**kwargs))

  def __exit__(self, *args):
    self.wait()
    if self.returncode == 0:
      with tarfile.open(mode='w', fileobj=self.tarfileobj) as tar:
        solution_dir = os.path.join(self.project_dir.name, self.project_name,
                                    self.solution_name)
        try:
          tar.add(os.path.join(solution_dir, 'syn/report'), arcname='report')
          tar.add(os.path.join(solution_dir, 'syn/verilog'), arcname='hdl')
          tar.add(os.path.join(solution_dir, self.cwd.name, 'vivado_hls.log'),
                  arcname='log/' + self.solution_name + '.log')
          for pattern in ('*.sched.adb.xml', '*.verbose.sched.rpt',
                          '*.verbose.sched.rpt.xml'):
            for f in glob.glob(
                os.path.join(solution_dir, '.autopilot', 'db', pattern)):
              tar.add(f, arcname='report/' + os.path.basename(f))
        except FileNotFoundError as e:
          self.returncode = 1
          _logger.error('%s', e)
    super().__exit__(*args)
    self.project_dir.cleanup()


XILINX_XML_NS = {'xd': 'http://www.xilinx.com/xd'}


def get_device_info(platform_path: str):
  """Extract device part number and target frequency from SDAccel platform.

  Currently only support 5.x platforms.

  Args:
    platform_path: Path to the platform directory, e.g.,
        '/opt/xilinx/platforms/xilinx_u200_qdma_201830_2'.

  Raises:
    ValueError: If cannot parse the platform properly.
  """
  device_name = os.path.basename(platform_path)
  try:
    platform_file = next(
        filter(os.path.isfile,
               (os.path.join(platform_path, 'hw', f'{device_name}.{ext}')
                for ext in ('xsa', 'dsa'))))
  except StopIteration as e:
    raise ValueError('cannot find platform file for %s' % device_name) from e
  with zipfile.ZipFile(platform_file) as platform:
    with platform.open(device_name + '.hpfm') as metadata:
      platform_info = ET.parse(metadata).find('./xd:component/xd:platformInfo',
                                              XILINX_XML_NS)
      if platform_info is None:
        raise ValueError('cannot parse platform')
      clock_period = platform_info.find(
          "./xd:systemClocks/xd:clock/[@xd:id='0']", XILINX_XML_NS)
      if clock_period is None:
        raise ValueError('cannot find clock period in platform')
      part_num = platform_info.find('xd:deviceInfo', XILINX_XML_NS)
      if part_num is None:
        raise ValueError('cannot find part number in platform')
      return {
          'clock_period':
              clock_period.attrib['{{{xd}}}period'.format(**XILINX_XML_NS)],
          'part_num':
              part_num.attrib['{{{xd}}}name'.format(**XILINX_XML_NS)]
      }


KERNEL_XML_TEMPLATE = r'''
<?xml version="1.0" encoding="UTF-8"?>
<root versionMajor="1" versionMinor="6">
  <kernel name="{top_name}" language="ip_c" vlnv="xilinx.com:RTLKernel:{top_name}:1.0" attributes="" preferredWorkGroupSizeMultiple="0" workGroupSize="1" interrupt="true">
    <ports>{ports}
    </ports>
    <args>{args}
    </args>
  </kernel>
</root>
'''

S_AXI_PORT = r'''
      <port name="s_axi_control" mode="slave" range="0x1000" dataWidth="32" portType="addressable" base="0x0"/>
'''

M_AXI_PORT_TEMPLATE = r'''
      <port name="m_axi_{name}" mode="master" range="0xFFFFFFFF" dataWidth="{width}" portType="addressable" base="0x0"/>
'''

AXIS_PORT_TEMPLATE = r'''
      <port name="{name}" mode="{mode}" dataWidth="{width}" portType="stream"/>
'''

ARG_TEMPLATE = r'''
      <arg name="{name}" addressQualifier="{addr_qualifier}" id="{arg_id}" port="{port_name}" size="{size:#x}" offset="{offset:#x}" hostOffset="0x0" hostSize="{host_size:#x}" type="{c_type}"/>
'''


def print_kernel_xml(top_name: str, axis_inputs: Iterable[Tuple[str, str,
                                                                ir.Type, str]],
                     axis_outputs: Iterable[Tuple[str, str, ir.Type,
                                                  str]], kernel_xml: TextIO):
  """Generate kernel.xml file.

  Args:
    top_name: Name of the top-level kernel function.
    axis_inputs: Sequence of (port_name, _, haoda_type, _) of input axis ports
    axis_outputs: Sequence of (port_name, _, haoda_type, _) of output axis ports
    kernel_xml: File object to write to.
  """
  ports = ''
  args = ''
  offset = 0
  arg_id = 0
  size = host_size = 8
  for mode, axis_ports in (('read_only', axis_inputs), ('write_only',
                                                        axis_outputs)):
    for port_name, _, haoda_type, _ in axis_ports:
      width = haoda_type.width_in_bits
      c_type = xml.sax.saxutils.escape('stream<ap_axiu<%d, 0, 0, 0>>&' % width)
      width += 8 + width // 8 * 2
      ports += AXIS_PORT_TEMPLATE.format(name=port_name, mode=mode,
                                         width=width).rstrip('\n')
      args += ARG_TEMPLATE.format(name=port_name,
                                  addr_qualifier=4,
                                  arg_id=arg_id,
                                  port_name=port_name,
                                  c_type=c_type,
                                  size=size,
                                  offset=offset,
                                  host_size=host_size).rstrip('\n')
    arg_id += 1
  kernel_xml.write(
      KERNEL_XML_TEMPLATE.format(top_name=top_name, ports=ports, args=args))


BRAM_FIFO_TEMPLATE = '''`default_nettype none

// first-word fall-through (FWFT) FIFO using block RAM
// based on HLS generated code
module {name} #(
  parameter MEM_STYLE  = "block",
  parameter DATA_WIDTH = {width},
  parameter ADDR_WIDTH = {addr_width},
  parameter DEPTH      = {depth}
) (
  input wire clk,
  input wire reset,

  // write
  output wire                  if_full_n,
  input  wire                  if_write_ce,
  input  wire                  if_write,
  input  wire [DATA_WIDTH-1:0] if_din,

  // read
  output wire                  if_empty_n,
  input  wire                  if_read_ce,
  input  wire                  if_read,
  output wire [DATA_WIDTH-1:0] if_dout
);

(* ram_style = MEM_STYLE *)
reg  [DATA_WIDTH-1:0] mem[0:DEPTH-1];
reg  [DATA_WIDTH-1:0] q_buf;
reg  [ADDR_WIDTH-1:0] waddr;
reg  [ADDR_WIDTH-1:0] raddr;
wire [ADDR_WIDTH-1:0] wnext;
wire [ADDR_WIDTH-1:0] rnext;
wire                  push;
wire                  pop;
reg  [ADDR_WIDTH-1:0] used;
reg                   full_n;
reg                   empty_n;
reg  [DATA_WIDTH-1:0] q_tmp;
reg                   show_ahead;
reg  [DATA_WIDTH-1:0] dout_buf;
reg                   dout_valid;

localparam DepthM1 = DEPTH[ADDR_WIDTH-1:0] - 1'd1;

assign if_full_n  = full_n;
assign if_empty_n = dout_valid;
assign if_dout    = dout_buf;
assign push       = full_n & if_write_ce & if_write;
assign pop        = empty_n & if_read_ce & (~dout_valid | if_read);
assign wnext      = !push              ? waddr              :
                    (waddr == DepthM1) ? {{ADDR_WIDTH{{1'b0}}}} : waddr + 1'd1;
assign rnext      = !pop               ? raddr              :
                    (raddr == DepthM1) ? {{ADDR_WIDTH{{1'b0}}}} : raddr + 1'd1;

// waddr
always @(posedge clk) begin
  if (reset)
    waddr <= {{ADDR_WIDTH{{1'b0}}}};
  else
    waddr <= wnext;
end

// raddr
always @(posedge clk) begin
  if (reset)
    raddr <= {{ADDR_WIDTH{{1'b0}}}};
  else
    raddr <= rnext;
end

// used
always @(posedge clk) begin
  if (reset)
    used <= {{ADDR_WIDTH{{1'b0}}}};
  else if (push && !pop)
    used <= used + 1'b1;
  else if (!push && pop)
    used <= used - 1'b1;
end

// full_n
always @(posedge clk) begin
  if (reset)
    full_n <= 1'b1;
  else if (push && !pop)
    full_n <= (used != DepthM1);
  else if (!push && pop)
    full_n <= 1'b1;
end

// empty_n
always @(posedge clk) begin
  if (reset)
    empty_n <= 1'b0;
  else if (push && !pop)
    empty_n <= 1'b1;
  else if (!push && pop)
    empty_n <= (used != {{{{(ADDR_WIDTH-1){{1'b0}}}},1'b1}});
end

// mem
always @(posedge clk) begin
  if (push)
    mem[waddr] <= if_din;
end

// q_buf
always @(posedge clk) begin
  q_buf <= mem[rnext];
end

// q_tmp
always @(posedge clk) begin
  if (reset)
    q_tmp <= {{DATA_WIDTH{{1'b0}}}};
  else if (push)
    q_tmp <= if_din;
end

// show_ahead
always @(posedge clk) begin
  if (reset)
    show_ahead <= 1'b0;
  else if (push && used == {{{{(ADDR_WIDTH-1){{1'b0}}}},pop}})
    show_ahead <= 1'b1;
  else
    show_ahead <= 1'b0;
end

// dout_buf
always @(posedge clk) begin
  if (reset)
    dout_buf <= {{DATA_WIDTH{{1'b0}}}};
  else if (pop)
    dout_buf <= show_ahead? q_tmp : q_buf;
end

// dout_valid
always @(posedge clk) begin
  if (reset)
    dout_valid <= 1'b0;
  else if (pop)
    dout_valid <= 1'b1;
  else if (if_read_ce & if_read)
    dout_valid <= 1'b0;
end

endmodule  // fifo_bram

`default_nettype wire
'''

SRL_FIFO_TEMPLATE = '''`default_nettype none

// first-word fall-through (FWFT) FIFO using shift register LUT
// based on HLS generated code
module {name} #(
  parameter MEM_STYLE  = "shiftreg",
  parameter DATA_WIDTH = {width},
  parameter ADDR_WIDTH = {addr_width},
  parameter DEPTH      = {depth}
) (
  input wire clk,
  input wire reset,

  // write
  output wire                  if_full_n,
  input  wire                  if_write_ce,
  input  wire                  if_write,
  input  wire [DATA_WIDTH-1:0] if_din,

  // read
  output wire                  if_empty_n,
  input  wire                  if_read_ce,
  input  wire                  if_read,
  output wire [DATA_WIDTH-1:0] if_dout
);

  wire [ADDR_WIDTH - 1:0] shift_reg_addr;
  wire [DATA_WIDTH - 1:0] shift_reg_data;
  wire [DATA_WIDTH - 1:0] shift_reg_q;
  wire                    shift_reg_ce;
  reg  [ADDR_WIDTH:0]     out_ptr;
  reg                     internal_empty_n;
  reg                     internal_full_n;

  reg [DATA_WIDTH-1:0] mem [0:DEPTH-1];

  assign if_empty_n = internal_empty_n;
  assign if_full_n = internal_full_n;
  assign shift_reg_data = if_din;
  assign if_dout = shift_reg_q;

  assign shift_reg_addr = out_ptr[ADDR_WIDTH] == 1'b0 ? out_ptr[ADDR_WIDTH-1:0] : {{ADDR_WIDTH{{1'b0}}}};
  assign shift_reg_ce = (if_write & if_write_ce) & internal_full_n;

  assign shift_reg_q = mem[shift_reg_addr];

  always @(posedge clk) begin
    if (reset) begin
      out_ptr <= ~{{ADDR_WIDTH+1{{1'b0}}}};
      internal_empty_n <= 1'b0;
      internal_full_n <= 1'b1;
    end else begin
      if (((if_read && if_read_ce) && internal_empty_n) &&
          (!(if_write && if_write_ce) || !internal_full_n)) begin
        out_ptr <= out_ptr - 1'b1;
        if (out_ptr == {{(ADDR_WIDTH+1){{1'b0}}}})
          internal_empty_n <= 1'b0;
        internal_full_n <= 1'b1;
      end
      else if (((if_read & if_read_ce) == 0 | internal_empty_n == 0) &&
        ((if_write & if_write_ce) == 1 & internal_full_n == 1))
      begin
        out_ptr <= out_ptr + 1'b1;
        internal_empty_n <= 1'b1;
        if (out_ptr == DEPTH - {{{{(ADDR_WIDTH-1){{1'b0}}}}, 2'd2}})
          internal_full_n <= 1'b0;
      end
    end
  end

  integer i;
  always @(posedge clk) begin
    if (shift_reg_ce) begin
      for (i = 0; i < DEPTH - 1; i = i + 1)
        mem[i + 1] <= mem[i];
      mem[0] <= shift_reg_data;
    end
  end

endmodule  // fifo_srl

`default_nettype wire
'''

AUTO_FIFO_TEMPLATE = '''`default_nettype none

// first-word fall-through (FWFT) FIFO
// if its capacity > 1024 bits, it uses block RAM, otherwise it will uses shift
// register LUT
module {name} #(
  parameter DATA_WIDTH = 32,
  parameter ADDR_WIDTH = 5,
  parameter DEPTH      = 32
) (
  input wire clk,
  input wire reset,

  // write
  output wire                  if_full_n,
  input  wire                  if_write_ce,
  input  wire                  if_write,
  input  wire [DATA_WIDTH-1:0] if_din,

  // read
  output wire                  if_empty_n,
  input  wire                  if_read_ce,
  input  wire                  if_read,
  output wire [DATA_WIDTH-1:0] if_dout
);

generate
  if (DATA_WIDTH * DEPTH > 1024) begin : bram
    fifo_bram #(
      .DATA_WIDTH(DATA_WIDTH),
      .ADDR_WIDTH(ADDR_WIDTH),
      .DEPTH     (DEPTH)
    ) unit (
      .clk  (clk),
      .reset(reset),

      .if_full_n  (if_full_n),
      .if_write_ce(if_write_ce),
      .if_write   (if_write),
      .if_din     (if_din),

      .if_empty_n(if_empty_n),
      .if_read_ce(if_read_ce),
      .if_read   (if_read),
      .if_dout   (if_dout)
    );
  end else begin : srl
    fifo_srl #(
      .DATA_WIDTH(DATA_WIDTH),
      .ADDR_WIDTH(ADDR_WIDTH),
      .DEPTH     (DEPTH)
    ) unit (
      .clk  (clk),
      .reset(reset),

      .if_full_n  (if_full_n),
      .if_write_ce(if_write_ce),
      .if_write   (if_write),
      .if_din     (if_din),

      .if_empty_n(if_empty_n),
      .if_read_ce(if_read_ce),
      .if_read   (if_read),
      .if_dout   (if_dout)
    );
  end
endgenerate

endmodule  // fifo

`default_nettype wire
'''


class VerilogPrinter(util.Printer):
  """A text-based Verilog printer."""

  def module(self, module_name: str, args: Iterable[str]) -> None:
    self.println('module %s (' % module_name)
    self.do_indent()
    self._out.write(' ' * self._indent * self._tab)
    self._out.write((',\n' + ' ' * self._indent * self._tab).join(args))
    self.un_indent()
    self.println('\n);')

  def endmodule(self, module_name: Optional[str] = None) -> None:
    if module_name is None:
      self.println('endmodule')
    else:
      self.println('endmodule // %s' % module_name)

  def begin(self) -> None:
    self.println('begin')
    self.do_indent()

  def end(self) -> None:
    self.un_indent()
    self.println('end')

  def parameter(self, key: str, value: str):
    self.println('parameter {} = {};'.format(key, value))

  @contextlib.contextmanager
  def initial(self) -> Iterator[None]:
    self.println('initial begin')
    self.do_indent()
    yield
    self.un_indent()
    self.println('end')

  @contextlib.contextmanager
  def always(self, condition: str) -> Iterator[None]:
    self.println('always @ (%s) begin' % condition)
    self.do_indent()
    yield
    self.un_indent()
    self.println('end')

  @contextlib.contextmanager
  def if_(self, condition: str) -> Iterator[None]:
    self.println('if (%s) begin' % condition)
    self.do_indent()
    yield
    self.end()

  def else_(self) -> None:
    self.un_indent()
    self.println('end else begin')
    self.do_indent()

  def module_instance(self, module_name: str, instance_name: str,
                      args: Union[Mapping[str, str], Iterable[str]]) -> None:
    self.println('{module_name} {instance_name}('.format(**locals()))
    self.do_indent()
    if isinstance(args, collections.Mapping):
      self._out.write(',\n'.join(' ' * self._indent * self._tab +
                                 '.{}({})'.format(*arg)
                                 for arg in args.items()))
    else:
      self._out.write(',\n'.join(
          ' ' * self._indent * self._tab + arg for arg in args))
    self.un_indent()
    self.println('\n);')

  def fifo_module(self,
                  width: int,
                  depth: int,
                  name: str = '',
                  threshold: int = 1024) -> None:
    """Generate FIFO with the given parameters.

    Generate an FIFO module. If its capacity is larger than threshold, BRAM FIFO
        will be used. Otherwise, SRL FIFO will be used.

    Args:
      width: FIFO width.
      depth: FIFO depth.
      name: Optionally give the fifo a name, default to
          'fifo_w{width}_d{depth}_A'.
      threshold: Optionally give a threshold to decide whether to use BRAM or
          SRL. Defaults to 1024 bits.

    Raises:
      ValueError: If depth or width is invalid.
    """
    if width * depth > threshold:
      self.bram_fifo_module(width, depth)
    else:
      self.srl_fifo_module(width, depth)

  def bram_fifo_module(self, width: int, depth: int, name: str = '') -> None:
    """Generate BRAM FIFO with the given parameters.

    Generate a BRAM FIFO module.

    Args:
      width: FIFO width.
      depth: FIFO depth.
      name: Optionally give the fifo a name, default to
          'fifo_w{width}_d{depth}_A'.

    Raises:
      ValueError: If depth or width is invalid.
    """
    if depth < 2:
      raise ValueError('Invalid BRAM FIFO depth: %d < 1' % depth)
    if not name:
      name = 'fifo_w{width}_d{depth}_A'.format(width=width, depth=depth)
    self._out.write(
        BRAM_FIFO_TEMPLATE.format(width=width,
                                  depth=depth,
                                  name=name,
                                  addr_width=(depth - 1).bit_length()))

  def srl_fifo_module(self, width: int, depth: int, name: str = '') -> None:
    """Generate SRL FIFO with the given parameters.

    Generate a SRL FIFO module.

    Args:
      width: FIFO width.
      depth: FIFO depth.
      name: Optionally give the fifo a name, default to
          'fifo_w{width}_d{depth}_A'.

    Raises:
      ValueError: If depth or width is invalid.
    """
    if depth < 2:
      raise ValueError('Invalid SRL FIFO depth: %d < 1' % depth)
    if not name:
      name = 'fifo_w{width}_d{depth}_A'.format(width=width, depth=depth)
    addr_width = (depth - 1).bit_length()
    self._out.write(
        SRL_FIFO_TEMPLATE.format(width=width,
                                 depth=depth,
                                 name=name,
                                 addr_width=addr_width,
                                 depth_width=addr_width + 1))



import contextlib
import logging
import os
import signal
from typing import (Any, Generator, Iterable, List, Optional, TextIO, Tuple,
                    TypeVar, Union)

T = TypeVar('T')

_logger = logging.getLogger().getChild(__name__)


class InternalError(Exception):
  pass


class SemanticError(Exception):
  pass


class SemanticWarn(Exception):
  pass


class InputError(Exception):
  pass


class Printer:
  """A text-based code printer."""

  def __init__(self, out: TextIO):
    self._out = out
    self._indent = 0
    self._assign = 0
    self._comments = []  # type: List[str]
    self._tab = 2

  def println(self, line: str = '', indent: int = -1) -> None:
    if indent < 0:
      indent = self._indent
    if line:
      self._out.write('%s%s\n' % (' ' * indent * self._tab, line))
    else:
      self._out.write('\n')

  def do_indent(self) -> None:
    self._indent += 1

  def un_indent(self) -> None:
    self._indent -= 1

  def do_scope(self, comment: str = '') -> None:
    self.println('{')
    self.do_indent()
    self._comments.append(comment)

  def un_scope(self, comment: str = '', suffix: str = '') -> None:
    self.un_indent()
    popped_comment = self._comments.pop()
    if comment:
      self.println('}%s // %s' % (suffix, comment))
    else:
      if popped_comment:
        self.println('}%s // %s' % (suffix, popped_comment))
      else:
        self.println('}%s' % suffix)


class CppPrinter(Printer):
  """A text-based C printer."""

  def new_var(self) -> str:
    self._assign += 1
    return self.last_var()

  def last_var(self, offset: int = -1) -> str:
    return 'assign_%d' % (self._assign + offset)

  def print_func(self,
                 name: str,
                 params: Iterable[str],
                 suffix: str = '',
                 align: int = 80) -> None:
    lines = [name + '(']
    for param in params:
      if ((self._indent + min(1,
                              len(lines) - 1)) * self._tab + len(lines[-1]) +
          len(param + ', ')) > align:
        lines.append(param + ', ')
      else:
        lines[-1] += param + ', '
    if lines[-1][-2:] == ', ':
      lines[-1] = lines[-1][:-2] + ')' + suffix
    line = lines.pop(0)
    self.println(line)
    if lines:
      self.do_indent()
      for line in lines:
        self.println(line)
      self.un_indent()

  @contextlib.contextmanager
  def for_(
      self, *args: Union[Tuple[str, str, str], Tuple[str, str]]
  ) -> Generator[None, None, None]:
    """Print a C++ for loop.

    Args:
      *args: 2 arguments for C++ 11 range-based for loop, 3 arguments for normal
          for loop.

    Raises:
        ValueError: If not given 2 or 3 arguments.
    """
    if len(args) == 3:
      self.println('for ({}; {}; {}) {{'.format(*args))
    elif len(args) == 2:
      self.println('for ({} : {}) {{'.format(*args))
    else:
      raise ValueError('for_ takes 2 or 3 arguments')
    self.do_indent()
    yield
    self.un_indent()
    self.println('}')

  @contextlib.contextmanager
  def do_while(self, cond: str) -> Generator[None, None, None]:
    self.println('do {')
    self.do_indent()
    yield
    self.un_indent()
    self.println('}} while ({});'.format(cond))

  @contextlib.contextmanager
  def if_(self, cond: str) -> Generator[None, None, None]:
    self.println('if ({}) {{'.format(cond))
    self.do_indent()
    yield
    self.un_indent()
    self.println('}')

  @contextlib.contextmanager
  def elif_(self, cond: str) -> Generator[None, None, None]:
    self.un_indent()
    self.println('}} else if ({}) {{'.format(cond))
    self.do_indent()
    yield

  @contextlib.contextmanager
  def else_(self) -> Generator[None, None, None]:
    self.un_indent()
    self.println('} else {')
    self.do_indent()
    yield


def print_define(printer: CppPrinter, var: str, val: str) -> None:
  printer.println('#ifndef %s' % var)
  printer.println('#define %s %s' % (var, val))
  printer.println('#endif  //%s' % var)


def print_guard(printer: CppPrinter, var: str, val: str) -> None:
  printer.println('#ifdef %s' % var)
  printer.println('#if %s != %s' % (var, val))
  printer.println('#error %s != %s' % (var, val))
  printer.println('#endif  //%s != %s' % (var, val))
  printer.println('#endif  //%s' % var)

  
def get_haoda_type(c_type: str) -> str:
  return c_type[:-2] if c_type[-2:] == '_t' else c_type


def get_suitable_int_type(upper: int, lower: int = 0) -> str:
  """Returns the suitable integer type with the least bits.

  Returns the integer type that can hold all values between max_val and min_val
  (inclusive) and has the least bits.

  Args:
    max_val: Maximum value that needs to be valid.
    min_val: Minimum value that needs to be valid.

  Returns:
    The suitable type.
  """
  assert upper >= lower
  upper = max(upper, 0)
  lower = min(lower, 0)
  if lower == 0:
    return 'uint%d' % upper.bit_length()
  return 'int%d' % (max(upper.bit_length(), (lower + 1).bit_length()) + 1)


def idx2str(idx: Iterable[Any]) -> str:
  return '(%s)' % ', '.join(map(str, idx))


def lst2str(idx: Iterable[Any]) -> str:
  return '[%s]' % ', '.join(map(str, idx))


def add_inv(idx: Iterable[int]) -> Tuple[int, ...]:
  return tuple(-x for x in idx)


def get_module_name(module_id: int) -> str:
  return 'module_%d' % module_id


def get_func_name(module_id: int) -> str:
  return 'Module%dFunc' % module_id


get_port_name = lambda name, bank: 'bank_{}_{}'.format(bank, name)
get_port_buf_name = lambda name, bank: 'bank_{}_{}_buf'.format(bank, name)


def get_bundle_name(name: str, bank: int):
  return '{}_bank_{}'.format(name.replace('<', '_').replace('>', ''), bank)


def pause_for_debugging() -> None:
  if _logger.isEnabledFor(logging.DEBUG):
    try:
      _logger.debug('pausing for debugging... send Ctrl-C to resume')
      signal.pause()
    except KeyboardInterrupt:
      pass


@contextlib.contextmanager
def timeout(seconds: int = 1, error_message: str = 'Timeout'):

  def handler(signum, frame):
    raise TimeoutError(error_message)

  signal.signal(signal.SIGALRM, handler)
  signal.alarm(seconds)
  yield
  signal.alarm(0)


def get_job_server_fd(
    job_server_fd: Union[int, Tuple[()], None]) -> Optional[int]:
  """Get the job server file descriptor from env var if input is a tuple.

  Args:
    job_server_fd: If this is not a tuple, return it directly.

  Returns:
    The job server file descriptor, or None.
  """
  if isinstance(job_server_fd, tuple):
    job_server = os.environ.get("JOB_SERVER_FD")
    if job_server is not None:
      job_server_fd = int(job_server)
    else:
      job_server_fd = None
  return job_server_fd


# pylint: disable=bad-whitespace
def acquire_job_slot(job_server_fd: Union[int, Tuple[()],
                                          None] = ()) -> Optional[int]:
  """Acquire a job slot if input is not None.

  Args:
    job_server_fd: If this is a tuple, obtain the fd from env var; if this is
        none, do nothing.

  Returns:
    The job server file descriptor, or None.
  """
  job_server_fd = get_job_server_fd(job_server_fd)
  if job_server_fd is not None and len(os.read(job_server_fd, 1)) != 1:
    job_server_fd = None
  return job_server_fd


# pylint: disable=bad-whitespace
def release_job_slot(job_server_fd: Union[int, Tuple[()],
                                          None] = ()) -> Optional[int]:
  """Release a job slot if input is not None.

  Args:
    job_server_fd: If this is a tuple, obtain the fd from env var; if this is
        none, do nothing.

  Returns:
    The job server file descriptor, or None.
  """
  job_server_fd = get_job_server_fd(job_server_fd)
  if job_server_fd is not None and os.write(job_server_fd, b'x') != 1:
    job_server_fd = None
  return job_server_fd



import collections
import functools
import logging
from collections import OrderedDict
from typing import (Callable, Dict, Iterable, List, Mapping, Tuple, TypeVar,
                    overload)

from haoda import ir, util

_logger = logging.getLogger().getChild(__name__)

NodeT = TypeVar('NodeT', bound=ir.Node)
T = TypeVar('T')

__all__ = ('simplify',)


# pylint: disable=function-redefined
@overload
def simplify(expr: NodeT, logger: Callable[..., None] = None) -> NodeT:
  ...


# pylint: disable=function-redefined
@overload
def simplify(expr: Iterable[NodeT], logger: Callable[..., None] = None) \
    -> Iterable[NodeT]:
  ...


def simplify(expr, logger=None):
  """Simplifies expressions.

  Args:
    expr: A haoda.NodeT or a sequence of haoda.ir.Node.

  Returns:
    Simplified haoda.ir.Node or sequence.
  """

  if expr is None:
    if logger is not None:
      logger.debug('None expr, no simplification.')
    return expr

  passes = flatten
  if logger is not None:
    passes = compose(passes, lambda node: print_tree(node, logger))

  if isinstance(expr, collections.Iterable):
    return type(expr)(map(passes, expr))

  return passes(expr)


def compose(*funcs: Callable[[T], T]) -> Callable[[T], T]:
  """Composes functions. The first function in funcs are invoked the first.
  """
  return functools.reduce(lambda g, f: lambda x: f(g(x)), funcs, lambda x: x)


def flatten(node: ir.Node) -> ir.Node:
  """Flattens an node if possible.

  Flattens an node if it is:
    + a singleton BinaryOp; or
    + a compound BinaryOp with reduction operators; or
    + a compound Operand; or
    + a Unary with an identity operator; or
    + a Call of a reduction functions.

  An Operand is a compound Operand if and only if its attr is a ir.Node.

  A Unary has identity operator if and only if all its operators are '+' or '-',
  and the number of '-' is even; or all of its operators are '!' and the number
  of '!' is even.

  Reduction functions are defined in ir.REDUCTION_FUNCS.

  Args:
    node: ir.Node to flatten.

  Returns:
    node: flattened ir.Node.

  Raises:
    util.InternalError: if Operand is undefined.
  """

  def visitor(node: ir.Node, args=None) -> ir.Node:
    if isinstance(node, ir.BinaryOp):

      # Flatten singleton BinaryOp
      if len(node.operand) == 1:
        return flatten(node.operand[0])

      # Flatten BinaryOp with reduction operators
      new_operator: List[str] = []
      new_operand: List[ir.Expr] = []
      for child_operator, child_operand in zip((None, *node.operator),
                                               node.operand):
        if child_operator is not None:
          new_operator.append(child_operator)
        # The first operator can always be flattened if two operations has the
        # same type.
        if child_operator in (None, '||', '&&', *'|&+*') and \
            type(child_operand) is type(node):
          new_operator.extend(child_operand.operator)
          new_operand.extend(child_operand.operand)
        else:
          new_operand.append(child_operand)
      # At least 1 operand is flattened.
      if len(new_operand) > len(node.operand):
        return flatten(type(node)(operator=new_operator, operand=new_operand))

    # Flatten compound Operand
    if isinstance(node, ir.Operand):
      for attr in node.ATTRS:
        val = getattr(node, attr)
        if val is not None:
          if isinstance(val, ir.Node):
            return flatten(val)
          break
      else:
        raise util.InternalError('undefined Operand')

    # Flatten identity unary operators
    if isinstance(node, ir.Unary):
      minus_count = node.operator.count('-')
      if minus_count % 2 == 0:
        plus_count = node.operator.count('+')
        if plus_count + minus_count == len(node.operator):
          return flatten(node.operand)
      not_count = node.operator.count('!')
      if not_count % 2 == 0 and not_count == len(node.operator):
        return flatten(node.operand)

    # Flatten reduction functions
    if isinstance(node, ir.Call):
      operator = getattr(node, 'name')
      if operator in ir.REDUCTION_FUNCS:
        operands: List[ir.Expr] = []
        for operand in getattr(node, 'arg'):
          if (isinstance(operand, ir.Call) and
              getattr(operand, 'name') == operator):
            operands.extend(getattr(operand, 'arg'))
          else:
            operands.append(operand)
        if len(operands) > len(getattr(node, 'arg')):
          return flatten(ir.Call(name=operator, arg=operands))

    return node

  if not isinstance(node, ir.Node):
    return node

  return node.visit(visitor)


def reverse_distribute(node: NodeT) -> NodeT:
  """Apply distributive property in reverse, if possible.

  Currently only left- and right-distribution of multiplication over addition is
  supported.

  Args:
    node: ir.Node to process.

  Returns:
    Processed node.
  """

  def visitor(node: NodeT, left_distribute: bool) -> NodeT:
    """Apply left- or right-distributive property in reverse, if possible

    Args:
      node: ir.Node to process.
      left_distribute: Whether to apply *left*-distributive property.

    Returns:
      Processed node.
    """
    if isinstance(node, ir.AddSub):
      items = OrderedDict()  # type: Dict[ir.Node, List[Tuple[str, ir.Node]]]
      new_operators = []
      new_operands = []
      for operator, operand in zip(('+',) + getattr(node, 'operator'),
                                   getattr(node, 'operand')):
        if (operator == '+' and isinstance(operand, ir.MulDiv) and
            getattr(operand, 'operator') == ('*',)):
          if left_distribute:
            coeff, item = getattr(operand, 'operand')
          else:
            item, coeff = getattr(operand, 'operand')
          items.setdefault(coeff, []).append((operator, item))
        else:
          new_operators.append(operator)
          new_operands.append(operand)
      for coeff, item in items.items():
        operator, operand = zip(*item)
        assert operator[0] == '+'
        new_operators.append(operator[0])
        if len(operand) > 1:
          new_item = ir.AddSub(operator=operator[1:], operand=operand)
        else:
          new_item = operand[0]
        if left_distribute:
          children = coeff, new_item
        else:
          children = new_item, coeff
        new_operands.append(ir.MulDiv(operator=('*',), operand=children))
      if len(new_operands) > 1:
        assert new_operators[0] == '+'
        new_node = ir.AddSub(operator=tuple(new_operators[1:]),
                             operand=tuple(new_operands))
        if new_node != node:
          return new_node  # type: ignore
      elif new_operands and new_operands[0] != node:
        return new_operands[0]
    return node

  return node.visit(visitor, True).visit(visitor, False)


def print_tree(node: NodeT,
               printer: Callable[[str], None] = _logger.debug) -> NodeT:
  """Prints the node as a tree.

  Args:
    node: ir.Node to print.
    printer: Optional, a callable that takes a string as input and print it,
        default to logging.Logger.debug.

  Returns:
    node: Input ir.Node as-is.
  """

  def pre_recursion(node: ir.Node, args: List[int]) -> None:
    args[0] += 1

  def post_recursion(node: ir.Node, args: List[int]) -> None:
    args[0] -= 1

  def visitor(node: NodeT, args: List[int]) -> NodeT:
    """Visitor that prints the node as a tree.

    Args:
      node: ir.Node to print.
      args: Singleton list of int, where the int is the current tree height.

    Returns:
      node: Input ir.Node as-is.
    """
    printer('%s+-%s(%s): %s' %
            (' ' * args[0], type(node).__name__, node.haoda_type, node))
    return node

  if not isinstance(node, ir.Node):
    return node

  printer('root')
  return node.visit(visitor,
                    args=[1],
                    pre_recursion=pre_recursion,
                    post_recursion=post_recursion)


def propagate_type(node: ir.Node, symbol_table: Mapping[str, ir.Type]):

  def visitor(node: ir.Node, symbol_table: Mapping[str, ir.Type]):
    if node.haoda_type is None:
      if isinstance(node, (ir.Ref, ir.Var)):
        node.haoda_type = symbol_table[node.name]
    return node

  return node.visit(visitor, symbol_table)



from typing import Any, Optional

from cached_property import cached_property

import haoda.util

TYPE_WIDTH = {'float': 32, 'double': 64, 'half': 16}
HAODA_TYPE_TO_CL_TYPE = {
    'uint8': 'uchar',
    'uint16': 'ushort',
    'uint32': 'uint',
    'uint64': 'ulong',
    'int8': 'char',
    'int16': 'short',
    'int32': 'int',
    'int64': 'long',
    'half': 'half',
    'float': 'float',
    'double': 'double',
    'float16': 'half',
    'float32': 'float',
    'float64': 'double',
}


class Type:

  def __init__(self, val: str):
    self._val = val

  def __str__(self) -> str:
    return self._val

  def __hash__(self) -> int:
    return self.width_in_bits

  def __eq__(self, other: Any) -> bool:
    if isinstance(other, str):
      other = Type(other)
    elif not isinstance(other, Type):
      return NotImplemented
    self_val = self._val
    other_val = other._val
    if self.is_float:
      self_val = 'float%d' % self.width_in_bits
    if other.is_float:
      other_val = 'float%d' % other.width_in_bits
    return self_val == other_val

  @cached_property
  def c_type(self) -> str:
    if self._val in {
        'uint8', 'uint16', 'uint32', 'uint64', 'int8', 'int16', 'int32', 'int64'
    }:
      return self._val + '_t'
    if self._val is None:
      return None
    if self._val == 'float32':
      return 'float'
    if self._val == 'float64':
      return 'double'
    for token in ('int', 'uint'):
      if self._val.startswith(token):
        bits = self._val.replace(token, '').split('_')
        if len(bits) > 1:
          assert len(bits) == 2
          return 'ap_{}<{}, {}>'.format(token.replace('int', 'fixed'), *bits)
        assert len(bits) == 1
        return 'ap_{}<{}>'.format(token, *bits)
    return self._val

  @cached_property
  def width_in_bits(self) -> int:
    if isinstance(self._val, str):
      if self._val in TYPE_WIDTH:
        return TYPE_WIDTH[self._val]
      for prefix in 'uint', 'int', 'float':
        if self._val.startswith(prefix):
          return int(self._val.lstrip(prefix).split('_')[0])
    else:
      if hasattr(self._val, 'haoda_type'):
        return self._val.haoda_type.width_in_bits
    raise haoda.util.InternalError('unknown haoda type: %s' % self._val)

  @cached_property
  def width_in_bytes(self) -> int:
    return (self.width_in_bits - 1) // 8 + 1

  def common_type(self, other: 'Type') -> 'Type':
    """Return the common type of two operands.

    TODO: Consider fractional.

    Args:
      lhs: Haoda type of operand 1.
      rhs: Haoda type of operand 2.

    Returns:
      The common type of two operands.
    """
    if self.is_float and not other.is_float:
      return self
    if other.is_float and not self.is_float:
      return other
    if self.width_in_bits < other.width_in_bits:
      return other
    return self

  @cached_property
  def is_float(self) -> bool:
    return self._val in {'half', 'double'} or self._val.startswith('float')

  @cached_property
  def is_fixed(self) -> bool:
    for token in ('int', 'uint'):
      if self._val.startswith(token):
        bits = self._val.replace(token, '').split('_')
        if len(bits) > 1:
          return True
    return False

  @cached_property
  def cl_type(self) -> Optional[str]:
    if self._val is None:
      return None
    cl_type = HAODA_TYPE_TO_CL_TYPE.get(self._val)
    if cl_type is not None:
      return cl_type
    return self._val + '_t'

  def get_cl_vec_type(self, burst_width: int) -> str:
    scalar_width = self.width_in_bits
    assert (burst_width % scalar_width == 0
           ), "burst width must be a multiple of width of the scalar type"
    assert (self._val in HAODA_TYPE_TO_CL_TYPE), "scalar type not supported"

    return HAODA_TYPE_TO_CL_TYPE[self._val] + str(burst_width // scalar_width)



import os
import re
import shutil
import tempfile
import xml.etree.ElementTree as ET
import zipfile
from typing import BinaryIO, Dict, Optional, TextIO

from haoda import util
from haoda.backend import xilinx as backend

REPORT_UTIL_COMMANDS = r'''
read_verilog [ glob {hdl_dir}/*.v ]
set ips [ glob -nocomplain {hdl_dir}/*/*.xci ]
if {{ $ips ne "" }} {{
  read_ip $ips
  generate_target synthesis [ get_files *.xci ]
}}
{set_parallel}
synth_design {synth_args}
report_utilization {report_util_args}
'''


class ReportXoUtil(backend.Vivado):
  """Run synthesis and generate resource utilization report.

  This class is a child of subprocess.Popen and will launch Vivado to run
  syntheis and generate resource utilization report. Arguments passed to
  synth_design and report_utilization can be configured via the kwargs params.

  Attributes:
    tmpdir: Temporary working directory for the RTL files and generated report.
    rpt_file: File object of generated resource utilization report.
    rpt_file_name: Name of the generated temporary resource utilization report.
    job_server_fd: Optional file descriptor of a job server.
    num_jobs: Optional number of jobs.
  """

  def __init__(self,
               xo_file: BinaryIO,
               rpt_file: TextIO,
               top_name: str = 'Dataflow',
               part_num: Optional[str] = None,
               synth_kwargs: Optional[Dict[str, str]] = None,
               report_util_kwargs: Optional[Dict[str, str]] = None):
    """Run synthesis and generate resource utilization report.

    Args:
      xo_file: XO file object containing the HDL files.
      rpt_file: File object of generated resource utilization report.
      top_name: Optionally specify a different top name other than "Dataflow".
      part_num: Optionally specify a different part number other than
          automatically determined.
      synth_kwargs: Dict of arguments for the synth_design command.
      report_util_kwargs: Dict of arguments for the synth_design command.

    Raises:
      InputError if input is not a valid XO.
    """
    self.tmpdir = tempfile.TemporaryDirectory(prefix='report-xo-util-')
    self.rpt_file = rpt_file
    self.rpt_file_name = os.path.join(self.tmpdir.name, 'post_synth_util.rpt')
    with zipfile.ZipFile(xo_file) as xo_zip:
      with xo_zip.open('xo.xml') as xo_xml:
        kernel = ET.parse(xo_xml).find('./Kernels/Kernel')
        if kernel is None:
          raise util.InputError('cannot parse XO file')
        ip_dir = kernel.attrib['IP']
        kernel_name = kernel.attrib['Name']
      hdl_dir_prefix = ip_dir + '/src'
      if all(not x.startswith(hdl_dir_prefix) for x in xo_zip.namelist()):
        hdl_dir_prefix = ip_dir + '/hdl/verilog'
      hdl_dir = os.path.join(self.tmpdir.name, hdl_dir_prefix)
      xo_zip.extractall(path=self.tmpdir.name,
                        members=[
                            name for name in xo_zip.namelist()
                            if name.startswith(hdl_dir_prefix)
                        ])
      if part_num is None:
        for hdl_file_format in ('{}.v', '{0}_{0}.v'):
          try:
            with open(os.path.join(
                hdl_dir, hdl_file_format.format(kernel_name))) as hdl_file:
              part_num = RtlHlsInfo(hdl_file)['HLS_INPUT_PART']
            break
          except FileNotFoundError:
            pass
    assert part_num is not None
    if synth_kwargs is None:
      synth_kwargs = {}
    if report_util_kwargs is None:
      report_util_kwargs = {}
    synth_kwargs.setdefault('top', top_name)
    synth_kwargs.setdefault('part', part_num)
    report_util_kwargs.setdefault('hierarchical', '')
    report_util_kwargs.setdefault('hierarchical_depth', '1')
    report_util_kwargs.setdefault('file', self.rpt_file_name)

    synth_args = ' '.join('-{} {}'.format(*kv) for kv in synth_kwargs.items())
    report_util_args = ' '.join(
        '-{} {}'.format(*kv) for kv in report_util_kwargs.items())
    kwargs = {
        'output_dir': os.path.join(self.tmpdir.name, 'output'),
        'hdl_dir': hdl_dir,
        'synth_args': synth_args,
        'report_util_args': report_util_args,
        'set_parallel': '',
    }
    self.job_server_fd = util.get_job_server_fd(())
    if self.job_server_fd is not None:
      new_fd = os.open('/proc/self/fd/%d' % self.job_server_fd,
                       os.O_RDWR | os.O_NONBLOCK)
      self.num_jobs = 1
      try:
        for i in range(7):
          self.num_jobs += len(os.read(new_fd, 1))
      except BlockingIOError as e:
        pass
      os.close(new_fd)
      kwargs['set_parallel'] = 'set_param general.maxThreads %d' % self.num_jobs
    super().__init__(REPORT_UTIL_COMMANDS.format(**kwargs))

  def __exit__(self, *args):
    super().__exit__(*args)
    if self.job_server_fd is not None:
      os.write(self.job_server_fd, b'x' * (self.num_jobs - 1))
    try:
      with open(self.rpt_file_name) as src_rpt_file:
        shutil.copyfileobj(src_rpt_file, self.rpt_file)
    except FileNotFoundError:
      raise util.InternalError('failed to generated report file')
    self.tmpdir.cleanup()


RTL_HLS_INFO_REGEX = r'\(\* CORE_GENERATION_INFO\s*=\s*".*,\{(.*)\}" \*\)'


class RtlHlsInfo:

  def __init__(self, rtl_file: TextIO):
    match = re.search(RTL_HLS_INFO_REGEX, rtl_file.read())
    if match is None:
      raise util.InputError('cannot parse RTL file')
    for item in match.group(1).split(','):
      key, value = item.split('=')
      setattr(self, key, value)

  def __getitem__(self, key: str) -> str:
    return getattr(self, key)



import collections
import copy
import logging
import math
from typing import List, Optional, Sequence, Tuple, Union

import cached_property

from haoda import ir, util
from haoda.ir import visitor

__all__ = (
    'AddSub',
    'BinaryAnd',
    'BinaryOp',
    'BinaryOr',
    'Call',
    'Cast',
    'DelayedRef',
    'DRAMRef',
    'EqCmp',
    'Expr',
    'FIFO',
    'FIFORef',
    'Let',
    'LogicAnd',
    'LtCmp',
    'Module',
    'ModuleTrait',
    'MulDiv',
    'Node',
    'Operand',
    'Ref',
    'Unary',
    'Var',
    'Xor',
    'from_reduction',
    'get_max_val',
    'get_min_val',
    'is_const',
    'make_var',
    'parenthesize',
    'str2int',
    'to_reduction',
    'unparenthesize',
    'FUNC_NAME',
    'FUNCS',
    'GRAMMAR',
    'MATH_FUNCS',
    'OTHER_FUNCS',
    'REDUCTION_FUNCS',
    'REDUCTION_OPS',
    'STD_FUNCS',
)

_logger = logging.getLogger().getChild(__name__)

MATH_FUNCS = ('cos', 'sin', 'tan', 'acos', 'asin', 'atan', 'atan2', 'cosh',
              'sinh', 'tanh', 'acosh', 'asinh', 'atanh', 'exp', 'frexp',
              'ldexp', 'log', 'log10', 'modf', 'exp2', 'expm1', 'ilogb',
              'log1p', 'log2', 'logb', 'scalbn', 'scalbln', 'pow', 'sqrt',
              'cbrt', 'hypot', 'erf', 'erfc', 'tgamma', 'lgamma', 'ceil',
              'floor', 'fmod', 'trunc', 'round', 'lround', 'llround', 'rint',
              'lrint', 'llrint', 'nearbyint', 'remainder', 'remquo', 'copysign',
              'nan', 'nextafter', 'nexttoward', 'fdim', 'fmax', 'fmin', 'fabs',
              'fma')

STD_FUNCS = ('abs', 'labs', 'llabs', 'div', 'ldiv', 'lldiv', 'imaxabs',
             'imaxdiv')

OTHER_FUNCS = ('min', 'max', 'select')

FUNCS = (tuple(map('/{}[fl]?/'.format, MATH_FUNCS)) +
         tuple(map("'{}'".format, STD_FUNCS + OTHER_FUNCS)))

FUNC_NAME = 'FuncName: %s;' % '|'.join(FUNCS)

GRAMMAR = r'''
Bin: /0[Bb][01]+([Uu][Ll][Ll]?|[Ll]?[Ll]?[Uu]?)/;
Dec: /\d+([Uu][Ll][Ll]?|[Ll]?[Ll]?[Uu]?)/;
Oct: /0[0-7]+([Uu][Ll][Ll]?|[Ll]?[Ll]?[Uu]?)/;
Hex: /0[Xx][0-9a-fA-F]+([Uu][Ll][Ll]?|[Ll]?[Ll]?[Uu]?)/;
Int: ('+'|'-')?(Hex|Bin|Oct|Dec);
Float: /(((\d*\.\d+|\d+\.)([+-]?[Ee]\d+)?)|(\d+[+-]?[Ee]\d+))[FfLl]?/;
Num: Float|Int;

Type: FixedType | FloatType;
FixedType: /u?int[1-9]\d*(_[1-9]\d*)?/;
FloatType: /float[1-9]\d*(_[1-9]\d*)?/ | 'float' | 'double' | 'half';

Let: (haoda_type=Type)? name=ID '=' expr=Expr;
Ref: name=ID '(' idx=INT (',' idx=INT)* ')' ('~' lat=Int)?;

Expr: operand=LogicAnd (operator=LogicOrOp operand=LogicAnd)*;
LogicOrOp: '||';

LogicAnd: operand=BinaryOr (operator=LogicAndOp operand=BinaryOr)*;
LogicAndOp: '&&';

BinaryOr: operand=Xor (operator=BinaryOrOp operand=Xor)*;
BinaryOrOp: '|';

Xor: operand=BinaryAnd (operator=XorOp operand=BinaryAnd)*;
XorOp: '^';

BinaryAnd: operand=EqCmp (operator=BinaryAndOp operand=EqCmp)*;
BinaryAndOp: '&';

EqCmp: operand=LtCmp (operator=EqCmpOp operand=LtCmp)*;
EqCmpOp: '=='|'!=';

LtCmp: operand=AddSub (operator=LtCmpOp operand=AddSub)*;
LtCmpOp: '<='|'>='|'<'|'>';

AddSub: operand=MulDiv (operator=AddSubOp operand=MulDiv)*;
AddSubOp: '+'|'-';

MulDiv: operand=Unary (operator=MulDivOp operand=Unary)*;
MulDivOp: '*'|'/'|'%';

Unary: (operator=UnaryOp)* operand=Operand;
UnaryOp: '+'|'-'|'~'|'!';

Operand: cast=Cast | call=Call | ref=Ref | num=Num | var=Var | '(' expr=Expr ')';
Cast: haoda_type=Type '(' expr=Expr ')';
Call: name=FuncName '(' arg=Expr (',' arg=Expr)* ')';
Var: name=ID ('[' idx=Int ']')*;

''' + FUNC_NAME

# pylint: disable=protected-access

class Node:
  """A immutable, hashable IR node.
  """
  SCALAR_ATTRS: Tuple[str, ...] = ()
  LINEAR_ATTRS: Tuple[str, ...] = ()

  @property
  def ATTRS(self) -> Tuple[str, ...]:
    return self.SCALAR_ATTRS + self.LINEAR_ATTRS

  def __init__(self, **kwargs):
    self._haoda_type = None
    for attr in self.SCALAR_ATTRS:
      setattr(self, attr, kwargs.pop(attr))
    for attr in self.LINEAR_ATTRS:
      setattr(self, attr, tuple(kwargs.pop(attr)))

  def __hash__(self) -> int:
    return hash((tuple(getattr(self, _) for _ in self.SCALAR_ATTRS),
                 tuple(tuple(getattr(self, _)) for _ in self.LINEAR_ATTRS)))

  def __eq__(self, other) -> bool:
    if (getattr(self, 'haoda_type', None) is not None and
        getattr(other, 'haoda_type', None) is not None and
        self.haoda_type != other.haoda_type):
      return False
    return all(
        hasattr(other, attr) and getattr(self, attr) == getattr(other, attr)
        for attr in self.ATTRS)

  @property
  def c_type(self) -> str:
    return self.haoda_type.c_type

  @property
  def cl_type(self) -> str:
    return self.haoda_type.cl_type

  def _get_haoda_type(self) -> ir.Type:
    """This method may be overridden by subclasses."""
    return self._haoda_type

  @property
  def haoda_type(self) -> ir.Type:
    return self._get_haoda_type()

  @haoda_type.setter
  def haoda_type(self, val: Union[None, str, ir.Type]) -> None:
    if val is None:
      self._haoda_type = None
    elif isinstance(val, str):
      self._haoda_type = ir.Type(val)
    elif isinstance(val, ir.Type):
      self._haoda_type = val
    else:
      raise ValueError(
          'haoda_type must be set from an instance of NoneType, str, '
          'or haoda.ir.Type, got %s' % type(val))

  @property
  def width_in_bits(self) -> int:
    return self.haoda_type.width_in_bits

  @property
  def c_expr(self) -> str:
    return self._get_expr('c')

  @property
  def cl_expr(self) -> str:
    return self._get_expr('cl')

  def _get_expr(self, lang: str) -> str:
    raise NotImplementedError

  def _get_type_expr(self, lang: str) -> str:
    if lang == 'c':
      return self.c_type
    if lang == 'cl':
      return self.cl_type
    raise ValueError(f'lang must be "c" or "cl", got "{lang}"')

  def visit(self, callback, args=None, pre_recursion=None, post_recursion=None):
    """A general-purpose, flexible, and powerful visitor.

    The args parameter will be passed to the callback callable so that it may
    read or write any information from or to the caller.

    A copy of self will be made and passed to the callback to avoid destructive
    access.

    If a new object is returned by the callback, it will be returned directly
    without recursion.

    If the same object is returned by the callback, if any attribute is
    changed, it will not be recursively visited. If an attribute is unchanged,
    it will be recursively visited.
    """

    def callback_wrapper(callback, obj, args):
      if callback is None:
        return obj
      result = callback(obj, args)
      if result is not None:
        return result
      return obj

    self_copy = copy.copy(self)
    obj = callback_wrapper(callback, self_copy, args)
    if obj is not self_copy:
      return obj
    self_copy = callback_wrapper(pre_recursion, copy.copy(self), args)
    scalar_attrs = {
        attr: getattr(self_copy, attr).visit(
            callback, args, pre_recursion, post_recursion) if isinstance(
                getattr(self_copy, attr), Node) else getattr(self_copy, attr)
        for attr in self_copy.SCALAR_ATTRS
    }
    linear_attrs = {
        attr: tuple(
            _.visit(callback, args, pre_recursion, post_recursion
                   ) if isinstance(_, Node) else _
            for _ in getattr(self_copy, attr))
        for attr in self_copy.LINEAR_ATTRS
    }

    for attr in self.SCALAR_ATTRS:
      # old attribute may not exist in mutated object
      if not hasattr(obj, attr):
        continue
      if getattr(obj, attr) is getattr(self, attr):
        if isinstance(getattr(obj, attr), Node):
          setattr(obj, attr, scalar_attrs[attr])
    for attr in self.LINEAR_ATTRS:
      # old attribute may not exist in mutated object
      if not hasattr(obj, attr):
        continue
      setattr(
          obj, attr,
          tuple(c if a is b and isinstance(a, Node) else a for a, b, c in zip(
              getattr(obj, attr), getattr(self, attr), linear_attrs[attr])))
    return callback_wrapper(post_recursion, obj, args)


class Let(Node):
  SCALAR_ATTRS = 'haoda_type', 'name', 'expr'

  name: str
  expr: Node

  def __str__(self):
    result = '{} = {}'.format(self.name, unparenthesize(self.expr))
    if self.haoda_type is not None:
      result = '{} {}'.format(self.haoda_type, result)
    return result

  def _get_haoda_type(self):
    if self._haoda_type is None:
      return self.expr.haoda_type
    return self._haoda_type

  def _get_expr(self, lang: str):
    return 'const {} {} = {};'.format(self._get_type_expr(lang), self.name,
                                      unparenthesize(self.expr._get_expr(lang)))


class Ref(Node):
  SCALAR_ATTRS = 'name', 'lat'
  LINEAR_ATTRS = ('idx',)

  name: str
  idx: Sequence[int]
  lat: Optional[int]

  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    self.idx = tuple(self.idx)
    if not hasattr(self, 'haoda_type'):
      self.haoda_type = None
    # self.lat will be defined in super().__init__(**kwargs)
    # pylint: disable=access-member-before-definition
    if isinstance(self.lat, str):
      self.lat = str2int(self.lat)

  def __str__(self):
    result = '{}({})'.format(self.name, ', '.join(map(str, self.idx)))
    if self.lat is not None:
      result += ' ~{}'.format(self.lat)
    return result


class BinaryOp(Node):
  LINEAR_ATTRS = 'operand', 'operator'

  operand: Sequence[Node]
  operator: Sequence[str]

  def __str__(self):
    result = str(self.operand[0])
    for operator, operand in zip(self.operator, self.operand[1:]):
      result += ' {} {}'.format(operator, operand)
    if self.singleton:
      return result
    return parenthesize(result)

  def _get_haoda_type(self):
    # TODO: derive from all operands
    return self.operand[0].haoda_type

  def _get_expr(self, lang: str) -> str:
    result = self.operand[0]._get_expr(lang)
    for operator, operand in zip(self.operator, self.operand[1:]):
      result += ' {} {}'.format(operator, operand._get_expr(lang))
    if self.singleton:
      return result
    return parenthesize(result)

  @property
  def singleton(self) -> bool:
    return len(self.operand) == 1  # type: ignore


class Expr(BinaryOp):
  pass


class LogicAnd(BinaryOp):
  pass


class BinaryOr(BinaryOp):
  pass


class Xor(BinaryOp):
  pass


class BinaryAnd(BinaryOp):
  pass


class EqCmp(BinaryOp):
  pass


class LtCmp(BinaryOp):
  pass


class AddSub(BinaryOp):
  '''
  def _get_haoda_type(self):
    if getattr(self, '_haoda_type', None) is None:

      # leave type undetermined if any child has undetermined type (None)
      for opd in self.operand:
        if opd.haoda_type is None:
          return None

      # use the biggest float if there is one
      float_types = {
          opd.haoda_type
          for opd in self.operand
          if opd.haoda_type.is_float
      }
      if float_types:
        for float_type, width in util.TYPE_WIDTH.items():
          if float_type in float_types:
            float_types.remove(float_type)
            float_types.add('float%d' % width)
        self._haoda_type = max(float_types, key=util.get_width_in_bits)
        return self._haoda_type

      # TODO: implement rules for fixed point non-interger numbers
      for opd in self.operand:
        if opd.haoda_type.is_fixed:
          self._haoda_type = opd.haoda_type
          return self._haoda_type

      # all operands are integers
      max_val = 0
      min_val = 0
      for opr, opd in zip(('+',) + self.operator, self.operand):
        if opr == '+':
          max_val += get_max_val(opd)
          min_val += get_min_val(opd)
        else:
          max_val -= get_min_val(opd)
          min_val -= get_max_val(opd)
      self._haoda_type = util.get_suitable_int_type(max_val, min_val)
    return self._haoda_type
  '''


class MulDiv(BinaryOp):
  '''
  @property
  def _get_haoda_type(self):
    if getattr(self, '_haoda_type', None) is None:

      # leave type undetermined if any child has undetermined type (None)
      for opd in self.operand:
        if opd.haoda_type is None:
          return None

      # use the biggest float if there is one
      float_types = {
          opd.haoda_type
          for opd in self.operand
          if opd.haoda_type.is_float
      }
      if float_types:
        for float_type, width in util.TYPE_WIDTH.items():
          if float_type in float_types:
            float_types.remove(float_type)
            float_types.add('float%d' % width)
        self._haoda_type = max(float_types, key=util.get_width_in_bits)
        return self._haoda_type

    # TODO: implement rules for fixed point non-interger numbers
      for opd in self.operand:
        if opd.haoda_type.is_fixed:
          self._haoda_type = opd.haoda_type
          return self._haoda_type

      # all operands are integers
      max_val = 1
      for opr, opd in zip(('*',) + self.operator, self.operand):
        if opr == '*':
          max_val *= max(abs(get_max_val(opd)), abs(get_min_val(opd)))
      self._haoda_type = util.get_suitable_int_type(max_val, -max_val)
    return self._haoda_type

  '''


class Unary(Node):
  SCALAR_ATTRS = ('operand',)
  LINEAR_ATTRS = ('operator',)

  operand: 'Operand'
  operator: Sequence[str]

  def __str__(self):
    return ''.join(self.operator) + str(self.operand)

  def _get_haoda_type(self):
    return self.operand.haoda_type

  def _get_expr(self, lang: str) -> str:
    return ''.join(self.operator) + self.operand._get_expr(lang)


class Operand(Node):
  SCALAR_ATTRS = 'cast', 'call', 'ref', 'num', 'var', 'expr'

  cast: Optional['Cast']
  call: Optional['Call']
  ref: Optional[Ref]
  num: Optional[str]
  var: Optional['Var']
  expr: Optional['Expr']

  def __str__(self):
    for attr in ('cast', 'call', 'ref', 'num', 'var'):
      if getattr(self, attr) is not None:
        return str(getattr(self, attr))
    # pylint: disable=useless-else-on-loop
    else:
      return parenthesize(self.expr)

  def _get_expr(self, lang: str) -> str:
    for attr_name in ('cast', 'call', 'ref', 'num', 'var'):
      attr = getattr(self, attr_name)
      if attr is not None:
        if hasattr(attr, '_get_expr'):
          return attr._get_expr(lang)
        return str(attr)
    # pylint: disable=useless-else-on-loop
    else:
      assert self.expr is not None
      return parenthesize(self.expr._get_expr(lang))

  def _get_haoda_type(self):
    for attr in self.ATTRS:
      val = getattr(self, attr)
      if val is not None:
        if hasattr(val, 'haoda_type'):
          return ir.Type(val.haoda_type)
        if attr == 'num':
          if 'u' in val.lower():
            if 'll' in val.lower():
              return ir.Type('uint64')
            return ir.Type('uint32')
          if 'll' in val.lower():
            return ir.Type('int64')
          if 'fl' in val.lower():
            return ir.Type('double')
          if 'f' in val.lower() or 'e' in val.lower():
            return ir.Type('float')
          if '.' in val:
            return ir.Type('double')
          return ir.Type('int32')
        return None
    raise util.InternalError('undefined Operand')


class Cast(Node):
  SCALAR_ATTRS = 'haoda_type', 'expr'

  expr: Node

  def __str__(self):
    return '{}{}'.format(self.haoda_type, parenthesize(self.expr))

  def _get_expr(self, lang: str) -> str:
    return '({}){}'.format(self._get_type_expr(lang),
                           parenthesize(self.expr._get_expr(lang)))


class Call(Node):
  SCALAR_ATTRS = ('name',)
  LINEAR_ATTRS = ('arg',)

  name: str
  arg: Sequence[Node]

  def __str__(self):
    return '{}({})'.format(self.name, ', '.join(map(str, self.arg)))

  def _get_haoda_type(self):
    if self.name in ('select',):
      if any(self.arg[idx].haoda_type is None for idx in (1, 2)):
        return None
      return util.common_type(self.arg[1].haoda_type, self.arg[2].haoda_type)
    return self.arg[0].haoda_type

  def _get_expr(self, lang: str) -> str:
    if self.name in {'min', 'max'}:
      assert len(self.arg) >= 2, 'too few arguments to %s' % self.name

      if lang == 'c':
        fmt_str = 'std::{}({}, {})'

        def variadic_to_binary_c(args: List[str]) -> str:
          nargs = len(args)
          if nargs == 1:
            return args[0]
          if nargs == 2:
            return fmt_str.format(self.name, *args)
          return fmt_str.format(self.name,
                                variadic_to_binary_c(args[:nargs // 2]),
                                variadic_to_binary_c(args[nargs // 2:]))

        return variadic_to_binary_c([_.c_expr for _ in self.arg])

      if lang == 'cl':
        fmt_str = '{}({}, {})'

        def variadic_to_binary_cl(args: Sequence[Node]) -> Tuple[str, bool]:
          nargs = len(args)
          func_name = self.name
          if nargs == 1:
            return args[0].cl_expr, args[0].haoda_type.is_float
          if nargs == 2:
            is_float = any(x.haoda_type.is_float for x in args)
            if is_float:
              func_name = f'f{func_name}'
            return fmt_str.format(func_name,
                                  *(x.cl_expr for x in args)), is_float
          arg1, is_float1 = variadic_to_binary_cl(args[:nargs // 2])
          arg2, is_float2 = variadic_to_binary_cl(args[nargs // 2:])
          return fmt_str.format(func_name, arg1, arg2), is_float1 or is_float2

        return variadic_to_binary_cl(self.arg)[0]

    if self.name == 'select':
      common_type = self.arg[1].haoda_type.common_type(self.arg[2].haoda_type)
      args = list(self.arg)
      for idx in 1, 2:
        if args[idx].haoda_type != common_type:
          args[idx] = Cast(haoda_type=common_type, expr=args[idx])
      return '({} ? {} : {})'.format(*(x._get_expr(lang) for x in args))
    return '{}({})'.format(self.name,
                           ', '.join(x._get_expr(lang) for x in self.arg))


class Var(Node):
  SCALAR_ATTRS = ('name',)
  LINEAR_ATTRS = ('idx',)

  name: str
  idx: Sequence[int]

  def __str__(self):
    return self.name + ''.join(map('[{}]'.format, self.idx))

  def _get_expr(self, lang: str) -> str:
    return self.name + ''.join(map('[{}]'.format, self.idx))


class FIFO(Node):
  """A reference to another node in a haoda.ir.Expr.

  This is used to represent a read/write from/to a Module in an output's Expr.
  It replaces Ref in haoda.ir, which is used to represent an element
  reference to a tensor.

  Attributes:
    read_module: Module reading from this FIFO.
    read_lat: int, at what cycle of a pipelined loop it is being read.
    write_module: Module writing to this FIFO.
    write_lat: int, at what cycle of a pipelined loop it is being written.
    depth: int, FIFO depth.
  """
  IMMUTABLE_ATTRS = 'read_module', 'write_module'
  SCALAR_ATTRS = 'read_module', 'read_lat', 'write_module', 'write_lat', 'depth'

  read_module: 'Module'
  read_lat: Optional[int]
  write_module: 'Module'
  write_lat: Optional[int]
  depth: Optional[int]

  def __init__(self,
               write_module,
               read_module,
               depth=None,
               write_lat=None,
               read_lat=None):
    super().__init__(write_module=write_module,
                     read_module=read_module,
                     depth=depth,
                     write_lat=write_lat,
                     read_lat=read_lat)

  def __repr__(self):
    return 'fifo[%d]: %s%s => %s%s' % (
        self.depth, repr(self.write_module), '' if self.write_lat is None else
        ' ~%s' % self.write_lat, repr(self.read_module),
        '' if self.read_lat is None else ' ~%s' % self.read_lat)

  def __hash__(self):
    return hash(tuple(getattr(self, _) for _ in self.IMMUTABLE_ATTRS))

  def __eq__(self, other):
    return all(
        getattr(self, _) == getattr(other, _)
        for _ in type(self).IMMUTABLE_ATTRS)

  @property
  def edge(self):
    return self.write_module, self.read_module

  def _get_haoda_type(self):
    return self.write_module.exprs[self].haoda_type

  def _get_expr(self, tag: str):
    return 'from_{}_to_{}'.format(self.write_module.name, self.read_module.name)


class Module():
  """A node in the dataflow graph.

  This is the base class for a dataflow module. It defines the parent (input)
  nodes, children (output) nodes, output expressions, input schedules, and
  output schedules. It also has a name to help identify itself.

  Attributes:
    parents: Set of parent (input) Module.
    children: Set of child (output) Module.
    lets: List of haoda.ir.Let expressions.
    exprs: Dict of {FIFO: haoda.ir.Expr}, stores an output's expression.
  """

  def __init__(self):
    """Initializes attributes into empty list or dict.
    """
    self.parents = []
    self.children = []
    self.lets = []
    self.exprs = collections.OrderedDict()

  @property
  def name(self):
    return 'module_%u' % hash(self)

  @property
  def fifos(self):
    return tuple(self.exprs.keys())

  @property
  def fifo_dict(self):
    return {(self, fifo.read_module): fifo for fifo in self.exprs}

  def fifo(self, dst_node):
    return self.fifo_dict[(self, dst_node)]

  def get_latency(self, dst_node):
    return self.fifo(dst_node).write_lat or 0

  def visit_loads(self, callback, args=None):
    obj = copy.copy(self)
    obj.lets = tuple(_.visit(callback, args) for _ in self.lets)
    obj.exprs = collections.OrderedDict()
    for fifo in self.exprs:
      obj.exprs[fifo] = self.exprs[fifo].visit(callback, args)
    return obj

  @property
  def dram_reads(self):
    return self._interfaces['dram_reads']

  @property
  def dram_writes(self):
    return self._interfaces['dram_writes']

  @property
  def input_fifos(self):
    return self._interfaces['input_fifos']

  @property
  def output_fifos(self):
    return self._interfaces['output_fifos']

  @cached_property.cached_property
  def _interfaces(self):
    # find dram reads
    reads_in_lets = tuple(_.expr for _ in self.lets)
    reads_in_exprs = tuple(self.exprs.values())
    dram_reads = collections.OrderedDict()
    for dram_ref in visitor.get_dram_refs(reads_in_lets + reads_in_exprs):
      for bank in dram_ref.dram:
        dram_reads[(dram_ref.var, bank)] = (dram_ref, bank)
    dram_reads = tuple(dram_reads.values())

    # find dram writes
    writes_in_lets = tuple(
        _.name for _ in self.lets if not isinstance(_.name, str))
    dram_writes = collections.OrderedDict()
    for dram_ref in visitor.get_dram_refs(writes_in_lets):
      for bank in dram_ref.dram:
        dram_writes[(dram_ref.var, bank)] = (dram_ref, bank)
    dram_writes = tuple(dram_writes.values())

    output_fifos = tuple(_.c_expr for _ in self.exprs)
    input_fifos = tuple(_.c_expr for _ in visitor.get_read_fifo_set(self))

    return {
        'dram_writes': dram_writes,
        'output_fifos': output_fifos,
        'input_fifos': input_fifos,
        'dram_reads': dram_reads
    }

  def __str__(self):
    return '%s @ 0x%x: %s' % (type(self).__name__, id(self), self.__dict__)

  def __repr__(self):
    return '%s @ 0x%x' % (type(self).__name__, id(self))

  def add_child(self, child):
    """Add a child (low level).

    This method only handles children and parents field; lets and exprs are
    not updated.

    Arguments:
      child: Module, child being added
    """
    if child not in self.children:
      self.children.append(child)
    if self not in child.parents:
      child.parents.append(self)

  def bfs_node_gen(self):
    """BFS over descendant nodes.

    This method is a BFS traversal generator over all descendant nodes.
    """
    node_queue = collections.deque([self])
    seen_nodes = {self}
    while node_queue:
      node = node_queue.popleft()
      yield node
      for child in node.children:
        if child not in seen_nodes:
          node_queue.append(child)
          seen_nodes.add(child)

  def dfs_node_gen(self):
    """DFS over descendant nodes.

    This method is a DFS traversal generator over all descendant nodes.
    """
    node_stack = [self]
    seen_nodes = {self}
    while node_stack:
      node = node_stack.pop()
      yield node
      for child in node.children:
        if child not in seen_nodes:
          node_stack.append(child)
          seen_nodes.add(child)

  def tpo_node_gen(self):
    """Traverse descendant nodes in topological order.

    This method is a generator that traverses all descendant nodes in
    topological order.
    """
    nodes = collections.OrderedDict()
    for node in self.bfs_node_gen():
      nodes[node] = len(node.parents)
    while nodes:
      for node in nodes:
        if nodes[node] == 0:
          yield node
          for child in node.children:
            nodes[child] -= 1
          del nodes[node]
          break
      else:
        return

  def bfs_edge_gen(self):
    """BFS over descendant edges.

    This method is a BFS traversal generator over all descendant edges.
    """
    node_queue = collections.deque([self])
    seen_nodes = {self}
    while node_queue:
      node = node_queue.popleft()
      for child in node.children:
        yield node, child
        if child not in seen_nodes:
          node_queue.append(child)
          seen_nodes.add(child)

  def dfs_edge_gen(self):
    """DFS over descendant edges.

    This method is a DFS traversal generator over all descendant edges.
    """
    node_stack = [self]
    seen_nodes = {self}
    while node_stack:
      node = node_stack.pop()
      for child in node.children:
        yield node, child
        if child not in seen_nodes:
          node_stack.append(child)
          seen_nodes.add(child)

  def get_descendants(self):
    """Get all descendant nodes.

    This method returns all descendant nodes as a set.

    Returns:
      Set of descendant Module.
    """
    return {self}.union(*map(Module.get_descendants, self.children))

  def get_connections(self):
    """Get all descendant edges.

    This method returns all descendant edges as a set.

    Returns:
      Set of descendant (src Module, dst Module) tuple.
    """
    return ({(self, child) for child in self.children
            }.union(*map(Module.get_connections, self.children)))


class DelayedRef(Node):
  """A delayed FIFO reference.

  Attributes:
    delay: int
    ref: FIFO
  """
  SCALAR_ATTRS = ('delay', 'ref')

  delay: int
  ref: Ref

  def _get_haoda_type(self):
    return self.ref.haoda_type

  def __str__(self):
    return '%s delayed %d' % (self.ref, self.delay)

  def __repr__(self):
    return str(self)

  def __hash__(self):
    return hash((self.delay, self.ref))

  def __eq__(self, other):
    return all(
        getattr(self, attr) == getattr(other, attr)
        for attr in ('delay', 'ref'))

  @property
  def buf_name(self):
    return '{ref.c_expr}_delayed_{delay}_buf'.format(**self.__dict__)

  @property
  def ptr(self):
    return '{ref.c_expr}_delayed_{delay}_ptr'.format(**self.__dict__)

  @property
  def ptr_type(self):
    return ir.Type('uint%d' % int(math.log2(self.delay) + 1))

  def _get_expr(self, lang: str) -> str:
    return '{}_delayed_{}'.format(self.ref._get_expr(lang), self.delay)

  @property
  def c_ptr_type(self):
    return self.ptr_type.c_type

  @property
  def cl_ptr_type(self):
    return self.ptr_type.cl_type

  @property
  def c_ptr_decl(self):
    return '{} {} = 0;'.format(self.c_ptr_type, self.ptr)

  @property
  def cl_ptr_decl(self):
    return '{} {} = 0;'.format(self.cl_ptr_type, self.ptr)

  @property
  def c_buf_ref(self):
    return '{}[{}]'.format(self.buf_name, self.ptr)

  @property
  def c_buf_decl(self):
    return '{} {}[{}];'.format(self.c_type, self.buf_name, self.delay)

  @property
  def cl_buf_decl(self):
    return '{} {}[{}];'.format(self.cl_type, self.buf_name, self.delay)

  @property
  def c_buf_load(self):
    return '{} = {};'.format(self.c_expr, self.c_buf_ref)

  @property
  def c_buf_store(self):
    return '{} = {};'.format(self.c_buf_ref, self.ref.ref_name)

  @property
  def c_next_ptr_expr(self):
    return '{ptr} < {depth} ? {c_ptr_type}({ptr}+1) : {c_ptr_type}(0)'.format(
        ptr=self.ptr, c_ptr_type=self.c_ptr_type, depth=self.delay - 1)

  @property
  def cl_next_ptr_expr(self):
    return '{ptr} < {depth} ? ({ptr_type})({ptr}+1) : ({ptr_type})0'.format(
        ptr=self.ptr, ptr_type=self.cl_ptr_type, depth=self.delay - 1)


class FIFORef(Node):
  """A FIFO reference.

  Attributes:
    fifo: FIFO it is linked to
    lat: int, at what cycle of a pipelined loop it is being referenced.
    ref_id: int, reference id in the current scope
  Properties:
    ld_name: str
    st_name: str
    ref_name: str
  """
  SCALAR_ATTRS = ('fifo', 'lat', 'ref_id')
  LD_PREFIX = 'fifo_ld_'
  ST_PREFIX = 'fifo_st_'
  REF_PREFIX = 'fifo_ref_'

  def __str__(self):
    return '<%s fifo_ref_%d%s>' % (self.haoda_type, self.ref_id,
                                   '@%s' % self.lat if self.lat else '')

  def __repr__(self):
    return str(self)

  def __hash__(self):
    return hash((self.lat, self.ref_id))

  def __eq__(self, other):
    if (self.haoda_type is not None and other.haoda_type is not None and
        self.haoda_type != other.haoda_type):
      return False
    return all(
        getattr(self, attr) == getattr(other, attr)
        for attr in ('lat', 'ref_id'))

  def _get_haoda_type(self):
    return self.fifo.haoda_type

  @property
  def ld_name(self):
    return '{}{}'.format(type(self).LD_PREFIX, self.ref_id)

  @property
  def ref_name(self):
    return '{}{}'.format(type(self).REF_PREFIX, self.ref_id)

  def _get_expr(self, lang: str) -> str:
    return self.ref_name


class DRAMRef(Node):
  """A DRAM reference.

  Attributes:
    haoda_type: str
    dram: [int], DRAM id it is accessing
    var: str, variable name it is accessing
    offset: int
  """
  SCALAR_ATTRS = 'haoda_type', 'dram', 'var', 'offset'

  def __str__(self):
    return 'dram<bank {} {}@{}>'.format(util.lst2str(self.dram), self.var,
                                        self.offset)

  def __repr__(self):
    return str(self)

  def __hash__(self):
    return hash((self.dram, self.offset))

  def __eq__(self, other):
    if (self.haoda_type is not None and other.haoda_type is not None and
        self.haoda_type != other.haoda_type):
      return False
    return all(
        getattr(self, attr) == getattr(other, attr)
        for attr in ('dram', 'offset'))

  def _get_expr(self, lang: str) -> str:
    return str(self)

  def dram_buf_name(self, bank):
    assert bank in self.dram, 'unexpected bank {}'.format(bank)
    return 'dram_{}_bank_{}_buf'.format(self.var, bank)

  def dram_fifo_name(self, bank):
    assert bank in self.dram, 'unexpected bank {}'.format(bank)
    return 'dram_{}_bank_{}_fifo'.format(self.var, bank)


class ModuleTrait(Node):
  """A immutable, hashable trait of a dataflow module.

  Attributes:
    lets: tuple of lets
    exprs: tuple of exprs
    template_types: tuple of template types (TODO)
    template_ints: tuple of template ints (TODO)

  Properties:
    loads: tuple of FIFORefs
  """
  LINEAR_ATTRS = ('lets', 'exprs', 'template_types', 'template_ints')

  def __init__(self, node):

    def mutate(obj, loads):
      if isinstance(obj, FIFO):
        if loads:
          if obj not in loads:
            load_id = next(reversed(loads.values())).ref_id + 1
          else:
            return loads[obj]
        else:
          load_id = 0
        fifo_ref = FIFORef(fifo=obj, lat=obj.read_lat, ref_id=load_id)
        loads[obj] = fifo_ref
        return fifo_ref
      return obj

    loads = collections.OrderedDict()
    node = node.visit_loads(mutate, loads)
    self.loads = tuple(loads.values())
    super().__init__(lets=tuple(node.lets),
                     exprs=tuple(node.exprs.values()),
                     template_types=tuple(),
                     template_ints=tuple())
    _logger.debug('Signature: %s', self)

  def __repr__(self):
    return '%s(loads: %s, lets: %s, exprs: %s)' % (
        type(self).__name__, util.idx2str(self.loads), util.idx2str(
            self.lets), util.idx2str(self.exprs))

  @property
  def dram_reads(self):
    return self._interfaces['dram_reads']

  @property
  def dram_writes(self):
    return self._interfaces['dram_writes']

  @property
  def input_fifos(self):
    return self._interfaces['input_fifos']

  @property
  def output_fifos(self):
    return self._interfaces['output_fifos']

  @cached_property.cached_property
  def _interfaces(self):
    # find dram reads
    reads_in_lets = tuple(_.expr for _ in self.lets)
    reads_in_exprs = tuple(self.exprs)
    dram_reads = collections.OrderedDict()
    for dram_ref in visitor.get_dram_refs(reads_in_lets + reads_in_exprs):
      for bank in dram_ref.dram:
        dram_reads[(dram_ref.var, bank)] = (dram_ref, bank)
    dram_reads = tuple(dram_reads.values())

    # find dram writes
    writes_in_lets = tuple(
        _.name for _ in self.lets if not isinstance(_.name, str))
    dram_writes = collections.OrderedDict()
    for dram_ref in visitor.get_dram_refs(writes_in_lets):
      for bank in dram_ref.dram:
        dram_writes[(dram_ref.var, bank)] = (dram_ref, bank)
    dram_writes = tuple(dram_writes.values())

    output_fifos = tuple('{}{}'.format(FIFORef.ST_PREFIX, idx)
                         for idx, expr in enumerate(self.exprs))
    input_fifos = tuple(_.ld_name for _ in self.loads)

    return {
        'dram_writes': dram_writes,
        'output_fifos': output_fifos,
        'input_fifos': input_fifos,
        'dram_reads': dram_reads
    }


def make_var(val):
  """Make literal Var from val."""
  return Var(name=val, idx=())


def is_const(node: Node) -> bool:
  if isinstance(node, Operand) and getattr(node, 'num') is not None:
    return True
  return False


def str2int(s, none_val=None):
  if s is None:
    return none_val
  while s[-1] in 'UuLl':
    s = s[:-1]
  if s[0:2] == '0x' or s[0:2] == '0X':
    return int(s, 16)
  if s[0:2] == '0b' or s[0:2] == '0B':
    return int(s, 2)
  if s[0] == '0':
    return int(s, 8)
  return int(s)


def parenthesize(expr) -> str:
  return '({})'.format(unparenthesize(expr))


def unparenthesize(expr) -> str:
  expr_str = str(expr)
  while expr_str.startswith('(') and expr_str.endswith(')'):
    count = 1
    for char in expr_str[1:-1]:
      if char == '(':
        count += 1
      elif char == ')':
        count -= 1
      if count == 0:  # the outermost parentheses are not paired
        return expr_str
    expr_str = expr_str[1:-1]
  return expr_str


def get_result_type(operand1, operand2, operator):
  for t in ('double', 'float') + sum(
      (('int%d_t' % w, 'uint%d_t' % w) for w in (64, 32, 16, 8)), tuple()):
    if t in (operand1, operand2):
      return t
  raise util.SemanticError('cannot parse type: %s %s %s' %
                           (operand1, operator, operand2))


def get_max_val(node) -> int:
  """Return the maximum valid value of an integer type."""
  if is_const(node):
    return int(node.num)
  haoda_type = node.haoda_type
  if haoda_type.is_float:
    raise TypeError('haoda_type has to be an integer type, got %s' % haoda_type)
  if haoda_type.is_fixed:
    raise NotImplementedError
  if haoda_type.startswith('uint'):
    return 2**haoda_type.width_in_bits - 1
  return 2**(haoda_type.width_in_bits - 1) - 1


def get_min_val(node) -> int:
  """Return the minimum valid value of an integer type."""
  if is_const(node):
    return int(node.num)
  haoda_type = node.haoda_type
  if haoda_type.is_float:
    raise TypeError('haoda_type has to be an integer type, got %s' % haoda_type)
  if haoda_type.is_fixed:
    raise NotImplementedError
  if haoda_type.startswith('uint'):
    return 0
  return -2**(haoda_type.width_in_bits - 1)


REDUCTION_OPS = {'+': AddSub, '*': MulDiv}
REDUCTION_FUNCS = {'min', 'max'}


def to_reduction(node: Node) -> Optional[Tuple[str, Tuple[Node, ...]]]:
  """Extract reduction expression from a Node.

  Args:
    node: Node to extract.

  Returns:
    None if node is not a reduction expression, tuple of (operator, operands)
    otherwise. operator is a string representing the reduction operator.
    operands is a tuple of operands as Nodes.
  """
  if isinstance(node, BinaryOp):
    operator = getattr(node, 'operator')
    if len(set(operator)) == 1 and operator[0] in REDUCTION_OPS:
      return operator[0], getattr(node, 'operand')
  elif isinstance(node, Call):
    operator = getattr(node, 'name')
    if operator in REDUCTION_FUNCS:
      return operator, getattr(node, 'arg')
  return None


def from_reduction(operator: str, operands: Tuple[Node, ...]) -> Node:
  """Assemble Node from a reduction expression.

  Args:
    operator: String representing the reduction operator.
    operands: Tuple of Nodes.

  Returns:
    Assembled Node.

  Raises:
    ValueError if operator is not a reduction operator.
  """
  if operator in REDUCTION_OPS:
    return REDUCTION_OPS[operator](operator=(operator,) * (len(operands) - 1),
                                   operand=operands)
  if operator in {'min', 'max'}:
    return Call(name=operator, arg=operands)
  raise ValueError('%s is not a reduction operator' % operator)



import collections

from haoda import ir


def get_read_fifo_set(module):
  """Get all read FIFOs as a tuple. Each FIFO only appears once.

  Args:
    module: A haoda.ir.Module object.

  Returns:
    A tuple of all FIFOs that are read in the module.

  Raises:
    TypeError: If argument is not a module.
  """

  def visitor(obj, args):
    if isinstance(obj, ir.FIFO):
      args[obj] = None
    return obj

  fifo_loads = collections.OrderedDict()
  if isinstance(module, ir.Module):
    module.visit_loads(visitor, fifo_loads)
  else:
    raise TypeError('argument is not a module')
  return tuple(fifo_loads)


def get_instances_of(node_or_iterable, class_or_tuple):
  """Get all ir.Node references of specific classes as a tuple.

  Args:
    node_or_iterable: A haoda.ir.Node object or an Iterable of
        haoda.ir.Node objects.
    class_or_tuple: Wanted class or tuple of wanted classes.

  Returns:
    A tuple of all wanted references.

  Raises:
    TypeError: If obj is not an IR node or a sequence.
  """

  def visitor(node, instances):
    if isinstance(node, class_or_tuple):
      instances.append(node)
    return node

  if isinstance(node_or_iterable, collections.Iterable):
    return sum(
        (get_instances_of(node, class_or_tuple) for node in node_or_iterable),
        ())
  instances = []
  if isinstance(node_or_iterable, ir.Node):
    node_or_iterable.visit(visitor, instances)
  else:
    raise TypeError('argument is not an IR node or a sequence')
  return tuple(instances)


def get_vars(node_or_iterable):
  return get_instances_of(node_or_iterable, ir.Var)


def get_dram_refs(node_or_iterable):
  return get_instances_of(node_or_iterable, ir.DRAMRef)



"""Hardware-Aware Optimization and Design Automation.

Reusable Python utilities for hardware-aware optimization and design automation.

See:
https://github.com/Blaok/haoda
"""

from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
  long_description = f.read()

setup(
    name='haoda',
    version='0.0.20200204.dev2',
    description='Hardware-aware optimization and design automation',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Blaok/haoda',
    author='Blaok Chi',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: System :: Hardware',
    ],
    packages=find_packages(exclude=('tests',)),
    python_requires='>=3.6',
    install_requires=['cached_property'],
)
